#include <time.h>
//Which print function is to use
#ifndef PROFILING
#	define my_print(...) Rprintf(__VA_ARGS__)
#else
#	define my_print(...) printf(__VA_ARGS__)
#endif

#define CHK_ERR() check_error(__LINE__);

//The name of the file which contains the OpenCL source code
#define CL_FILE_NAME "AlignAndTranslate.cl"

#include <CL/cl.h>	//OpenCL
#include <R.h>
#include <Rinternals.h>

//For the reading of the OpenCL source code from a file
#include <stdio.h>
#include <errno.h>
#include <string.h>

void check_error(int line);
const char *get_error_string(cl_int err);
void print_error(int line, const char *err_msg);
char *read_file(char *file_name);

void align_on_GPU(	int no_of_sequences, char sequences[], int starts_at[],
					float match_value, float mismatch_value,
					float gap_opening, float gap_extension,
					int result[]);

void align_and_translate_on_GPU(
					int no_of_sequences, char sequences[], int starts_at[],
					float match_value, float mismatch_value,
					float gap_opening, float gap_extension,
					char result[]);

cl_int g_err;
cl_platform_id g_platform;
cl_device_id g_device;
cl_context g_context;
cl_command_queue g_command_queue;
cl_program g_program;

//Executed when library is loaded by R
//Initializes ressources needed by OpenCL
//For the creation of the program the path of the source file is needed
void init(char *package_path[])
{
	//Find available platforms
	cl_uint no_of_platforms;
	g_err = clGetPlatformIDs(0, NULL, &no_of_platforms);
	if (no_of_platforms == 0){
		error("Couldn't find any OpenCL platform");
	}
	CHK_ERR();
	cl_platform_id *platforms = (cl_platform_id *)
				malloc(no_of_platforms * sizeof(cl_platform_id));
	g_err = clGetPlatformIDs(no_of_platforms, platforms, NULL);
	CHK_ERR();
	
	//Look for a GPU
	_Bool found_gpu = FALSE;	
	for (int i = 0; !found_gpu && i < no_of_platforms; i++) {
		cl_uint no_of_gpus;
		clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &no_of_gpus);
		if (no_of_gpus != 0) {
			g_platform = platforms[i];
			found_gpu = TRUE;
		}
	}
	if (!found_gpu) {
		my_print("Couldn't find any GPU. The first OpenCL platform will be used.\n");
		g_platform = platforms[0];
	}
	
	free(platforms);
	
	//Get device
	g_err = clGetDeviceIDs(g_platform, CL_DEVICE_TYPE_ALL, 1, &g_device, NULL);
	CHK_ERR();
	
	//Create context
	g_context = clCreateContext(NULL, 1, &g_device, NULL, NULL, &g_err);
	CHK_ERR();
	
	//Create command queue
	g_command_queue = clCreateCommandQueue(g_context, g_device, 0, &g_err);
	CHK_ERR();
	
	//Create Program
	char file_path[1000];
	const char* file_content;
	
	file_path[0] = '\0';
	#ifndef PROFILING
	strcat(file_path, package_path[0]);
	#else
	strcat(file_path, "../inst");
	#endif
	strcat(file_path, "/OpenCL/");
	strcat(file_path, CL_FILE_NAME);
	file_content = read_file(file_path);
	
	g_program = clCreateProgramWithSource(g_context, 1, &file_content,
										NULL, &g_err);
	CHK_ERR();
	
	//Compile OpenCL source code and print it, if there are errors
	g_err = clBuildProgram(g_program, 0, NULL, NULL, NULL, NULL);
	CHK_ERR();
	if (g_err != CL_SUCCESS) {
		char *build_log;
		size_t build_log_size;
		g_err = clGetProgramBuildInfo(g_program, g_device, CL_PROGRAM_BUILD_LOG,
										0, NULL, &build_log_size);
		CHK_ERR();
		build_log = (char *)malloc(build_log_size);
		g_err = clGetProgramBuildInfo(g_program, g_device, CL_PROGRAM_BUILD_LOG,
										build_log_size, build_log, NULL);
		CHK_ERR();
		my_print("%s\n", build_log);
		free((void *)build_log);
	}
	
	free((void *)file_content);
}

//Executed when library is unloaded by R
//Deletes all created OpenCL objects
void R_unload_ClassifyOnGPU()
{
	g_err = clReleaseCommandQueue(g_command_queue);
	CHK_ERR();
	
	g_err = clRetainProgram(g_program);
	CHK_ERR();
	
	g_err = clReleaseContext(g_context);
	CHK_ERR();
}

SEXP call_align_on_GPU(SEXP reference, SEXP subjects,
			SEXP match_value, SEXP mismatch_value,
			SEXP gap_opening, SEXP gap_extension)
{
	int i, len;
	int no_of_sequences = length(subjects) + 1;//+1 for reference
	//Save all sequences in one string
	//Where the sequences in the concatenated string starts
	//Calculated here and not in the concatenation, because the size of all the sequences is needed to be known to be able to allocate the memory for the concatened string
	int *starts_at = (int *)malloc(sizeof(int) * (no_of_sequences + 1));
	char *all_sequences, *copy_to;
	starts_at[0] = 0;
	starts_at[1] = strlen(CHAR(STRING_ELT(reference, 0)));
	for (i = 0; i < no_of_sequences - 1; i++){
		//starts_at[i + 2] = next sequences start
		starts_at[i + 2] = starts_at[i + 1] + strlen(CHAR(STRING_ELT(subjects, i)));
	}
	
	//Concatenate
	all_sequences = (char *)malloc(sizeof(char) * starts_at[no_of_sequences]);
	copy_to = all_sequences;
	len = starts_at[1] - starts_at[0];
	strncpy(copy_to, CHAR(STRING_ELT(reference, 0)), len);
	copy_to += len;
	for(i = 0; i < no_of_sequences - 1; i++){
		len = starts_at[i + 2] - starts_at[i + 1];
		strncpy(copy_to, CHAR(STRING_ELT(subjects, i)), len);
		copy_to += len;
	}
	
	SEXP result;
	PROTECT(result = allocVector(INTSXP, 2 * (no_of_sequences - 1)));
	
	align_on_GPU(
					no_of_sequences, all_sequences, starts_at,
					*REAL(match_value), *REAL(mismatch_value),
					*REAL(gap_opening), *REAL(gap_extension),
					INTEGER(result)
					);
	
	UNPROTECT(1);//Give control over result to the garbage collector
	return result;
}

SEXP call_align_and_translate_on_GPU(SEXP reference, SEXP subjects,
			SEXP match_value, SEXP mismatch_value,
			SEXP gap_opening, SEXP gap_extension)
{
	int i, len;
	int no_of_sequences = length(subjects) + 1;//+1 for reference
	//Save all sequences in one string
	//Where the sequences in the concatenated string starts
	//Calculated here and not in the concatenation, because the size of all the sequences is needed to be known to be able to allocate the memory for the concatened string
	int *starts_at = (int *)malloc(sizeof(int) * (no_of_sequences + 1));
	char *all_sequences, *copy_to;
	starts_at[0] = 0;
	starts_at[1] = strlen(CHAR(STRING_ELT(reference, 0)));
	for (i = 0; i < no_of_sequences - 1; i++){
		//starts_at[i + 2] = next sequences start
		starts_at[i + 2] = starts_at[i + 1] + strlen(CHAR(STRING_ELT(subjects, i)));
	}
	
	//Concatenate
	all_sequences = (char *)malloc(sizeof(char) * starts_at[no_of_sequences]);
	copy_to = all_sequences;
	len = starts_at[1] - starts_at[0];
	strncpy(copy_to, CHAR(STRING_ELT(reference, 0)), len);
	copy_to += len;
	for(i = 0; i < no_of_sequences - 1; i++){
		len = starts_at[i + 2] - starts_at[i + 1];
		strncpy(copy_to, CHAR(STRING_ELT(subjects, i)), len);
		copy_to += len;
	}
	
	char *translated = (char *)malloc((starts_at[no_of_sequences] - starts_at[1]) / 3
						+ no_of_sequences - 1);

	align_and_translate_on_GPU(
					no_of_sequences, all_sequences, starts_at,
					*REAL(match_value), *REAL(mismatch_value),
					*REAL(gap_opening), *REAL(gap_extension),
					translated
					);
	SEXP result;
	PROTECT(result = allocVector(STRSXP, no_of_sequences - 1));
	for(int i = 0; i < no_of_sequences - 1; i++) {
		SET_STRING_ELT(result, i, mkChar(
			translated + (starts_at[i + 1] - starts_at[1]) / 3 + i));
	}
	free(translated);
	UNPROTECT(1);//Give control over result to the garbage collector
	return result;
}



#define VECTOR_SIZE 3
#define ALIGN_MAX_WG 1024
void align_on_GPU(	int no_of_sequences, char sequences[], int starts_at[],
					float match_value, float mismatch_value,
					float gap_opening, float gap_extension,
					int result[])
{
	cl_kernel kernel;
	cl_mem 	sequences_mem, starts_at_mem, result_mem;
	cl_mem	a_value_mem, a_start_mem, b_value_mem, b_start_mem;
	size_t work_items_per_work_group = 160;
	size_t work_groups = ceil((no_of_sequences - 1) / (float)work_items_per_work_group);
	if (work_groups > ALIGN_MAX_WG)
		work_groups = ALIGN_MAX_WG;
	size_t total_work_items = work_groups * work_items_per_work_group;
	
	//Allocate Memory
	sequences_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					ceil(starts_at[no_of_sequences] /4.0) * 4.0 * sizeof(char), sequences, &g_err);
	CHK_ERR();
	
	starts_at_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					(no_of_sequences + 1) * sizeof(int), starts_at, &g_err);
	CHK_ERR();
	
	//Space for a_value, a_start, b_value, b_start. Each work-item needs strlen(ref_seq) + 1 cells
	int buffer_size = total_work_items * (starts_at[1] + 1);
	a_value_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(float), NULL, &g_err);
	CHK_ERR();
	a_start_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(int), NULL, &g_err);
	CHK_ERR();
	b_value_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(float), NULL, &g_err);
	CHK_ERR();
	b_start_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(int), NULL, &g_err);
	CHK_ERR();
	
	result_mem =	clCreateBuffer(g_context, CL_MEM_WRITE_ONLY,
					(no_of_sequences - 1) * 2 * sizeof(int), NULL, &g_err);
	CHK_ERR();

	//Create Kernel
	kernel = clCreateKernel(g_program, "align", &g_err);
	CHK_ERR();
	
	//Set parameters
	int param = 0;
	g_err  = clSetKernelArg(kernel, param++, sizeof(cl_mem), &sequences_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &starts_at_mem);
	int no_of_subjects = no_of_sequences - 1;
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_subjects);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &match_value);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &mismatch_value);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &gap_opening);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &gap_extension);
	
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &a_value_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &a_start_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &b_value_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &b_start_mem);
	
	g_err |= clSetKernelArg(kernel, param++, (VECTOR_SIZE + 1) * work_items_per_work_group * sizeof(float), NULL);
	g_err |= clSetKernelArg(kernel, param++, (VECTOR_SIZE + 1) * work_items_per_work_group * sizeof(int), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(float), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(int), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(char), NULL);
	
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &result_mem);

	CHK_ERR();
	
	//Run
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
			&total_work_items, &work_items_per_work_group, 0, NULL, NULL);
	CHK_ERR();
	
	//Load result	
	g_err = clEnqueueReadBuffer(g_command_queue, result_mem, CL_TRUE, 0,
			(no_of_sequences - 1) * 2 * sizeof(int), result, 0, NULL, NULL);
			
	//Clean up
	g_err  = clReleaseKernel(kernel);
	g_err |= clReleaseMemObject(sequences_mem);
	g_err |= clReleaseMemObject(starts_at_mem);
	g_err |= clReleaseMemObject(a_value_mem);
	g_err |= clReleaseMemObject(a_start_mem);
	g_err |= clReleaseMemObject(b_value_mem);
	g_err |= clReleaseMemObject(b_start_mem);
	g_err |= clReleaseMemObject(result_mem);
	CHK_ERR();
}

void align_and_translate_on_GPU(	int no_of_sequences, char sequences[], int starts_at[],
					float match_value, float mismatch_value,
					float gap_opening, float gap_extension,
					char result[])
{
	cl_kernel kernel;
	cl_mem 	sequences_mem, starts_at_mem, result_mem;
	cl_mem	a_value_mem, a_start_mem, b_value_mem, b_start_mem;
	cl_mem translated_mem;
	
	size_t work_items_per_work_group = 160;
	size_t work_groups = ceil((no_of_sequences - 1) / (float)work_items_per_work_group);
	if (work_groups > ALIGN_MAX_WG)
		work_groups = ALIGN_MAX_WG;
	size_t total_work_items = work_groups * work_items_per_work_group;
	
	//Allocate Memory
	sequences_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					ceil(starts_at[no_of_sequences] /4.0) * 4.0 * sizeof(char), sequences, &g_err);
	CHK_ERR();
	
	starts_at_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					(no_of_sequences + 1) * sizeof(int), starts_at, &g_err);
	CHK_ERR();
	
	//Space for a_value, a_start, b_value, b_start. Each work-item needs strlen(ref_seq) + 1 cells
	int buffer_size = total_work_items * (starts_at[1] + 1);
	a_value_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(float), NULL, &g_err);
	CHK_ERR();
	a_start_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(int), NULL, &g_err);
	CHK_ERR();
	b_value_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(float), NULL, &g_err);
	CHK_ERR();
	b_start_mem =	clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					buffer_size * sizeof(int), NULL, &g_err);
	CHK_ERR();
	
	result_mem =	clCreateBuffer(g_context, CL_MEM_WRITE_ONLY,
					(no_of_sequences - 1) * 2 * sizeof(int), NULL, &g_err);
	CHK_ERR();

	//Create Kernel
	kernel = clCreateKernel(g_program, "align", &g_err);
	CHK_ERR();
	
	//Set parameters
	int param = 0;
	g_err  = clSetKernelArg(kernel, param++, sizeof(cl_mem), &sequences_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &starts_at_mem);
	int no_of_subjects = no_of_sequences - 1;
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_subjects);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &match_value);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &mismatch_value);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &gap_opening);
	g_err |= clSetKernelArg(kernel, param++, sizeof(float), &gap_extension);
	
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &a_value_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &a_start_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &b_value_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &b_start_mem);
	
	g_err |= clSetKernelArg(kernel, param++, (VECTOR_SIZE + 1) * work_items_per_work_group * sizeof(float), NULL);
	g_err |= clSetKernelArg(kernel, param++, (VECTOR_SIZE + 1) * work_items_per_work_group * sizeof(int), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(float), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(int), NULL);
	g_err |= clSetKernelArg(kernel, param++, VECTOR_SIZE * work_items_per_work_group * sizeof(char), NULL);
	
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &result_mem);

	CHK_ERR();
	
	//Run
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
			&total_work_items, &work_items_per_work_group, 0, NULL, NULL);
	CHK_ERR();
	
	//Free memory
	g_err  = clReleaseKernel(kernel);	
	g_err |= clReleaseMemObject(a_value_mem);
	g_err |= clReleaseMemObject(a_start_mem);
	g_err |= clReleaseMemObject(b_value_mem);
	g_err |= clReleaseMemObject(b_start_mem);
	CHK_ERR();
	
	//Translate	
	//Allocate memory for result
	translated_mem = clCreateBuffer(g_context, CL_MEM_WRITE_ONLY,
				(starts_at[no_of_sequences] - starts_at[1]) / 3 + no_of_sequences - 1,
				NULL, &g_err);
	CHK_ERR();
	
	//Create Kernel
	kernel = clCreateKernel(g_program, "translate", &g_err);
	CHK_ERR();
	
	//Set parameters
	param = 0;
	g_err  = clSetKernelArg(kernel, param++, sizeof(cl_mem), &sequences_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &starts_at_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_subjects);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &result_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &translated_mem);
	CHK_ERR();
	#define MEASURE_TRANSLATION	0
	//RUN
	work_items_per_work_group = 192;
	total_work_items = 1024 * work_items_per_work_group;
	#if MEASURE_TRANSLATION
	struct timespec start, end;
	clFinish(g_command_queue);
	clock_gettime(CLOCK_MONOTONIC, &start);
	for(int loop = 0; loop < 50; loop++){
	#endif
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
			&total_work_items, &work_items_per_work_group, 0, NULL, NULL);
	CHK_ERR();
	#if MEASURE_TRANSLATION
	clFinish(g_command_queue);
	}	
	clock_gettime(CLOCK_MONOTONIC, &end);
	double secs = end.tv_sec - start.tv_sec + (end.tv_nsec - start.tv_nsec) / 1000000000.0;
	my_print("%f\n", secs/50.0f);
	#endif
	
	//Load result	
	g_err = clEnqueueReadBuffer(g_command_queue, translated_mem, CL_TRUE, 0,
			(starts_at[no_of_sequences] - starts_at[1]) / 3 + no_of_sequences - 1,
			result, 0, NULL, NULL);
	CHK_ERR();
	
	//Clean up
	g_err  = clReleaseKernel(kernel);
	g_err |= clReleaseMemObject(sequences_mem);
	g_err |= clReleaseMemObject(starts_at_mem);
	g_err |= clReleaseMemObject(result_mem);
	g_err |= clReleaseMemObject(translated_mem);
	CHK_ERR();
}

//Returns the file content as a string
//Do not forget to use free
//Approach from http://www.cplusplus.com/reference/cstdio/fread/
char *read_file(char *file_name)
{
	FILE *file;
	long file_length;
	char *file_content;
	size_t read_amount;
	
	//Open file
	file = fopen(file_name, "r");
	if (file == NULL){
		print_error(__LINE__ - 2, strerror(errno));
		return NULL;
	}
	
	//Determine file length
	fseek(file, 0, SEEK_END);//Go to the end of the file
	file_length = ftell(file);//Read position in file
	rewind(file);//Go to the beginning of the file
	
	//Allocate memory
	file_content = (char *)malloc(file_length * sizeof(char));
	
	//Read content
	read_amount = fread(file_content, sizeof(char), file_length, file);
	if (read_amount != file_length) {
		print_error(__LINE__ - 2, "Read less than should");
		return NULL;
	}
	
	//The string must be null-terminated, so it's length can be determined
	file_content[file_length - 1] = '\0';
	//Close file
	fclose(file);
	return file_content;	
}

//Prints an error message
void print_error(int line, const char *err_msg)
{
	my_print("Error in line %d: %s\n", line, err_msg);
}

//Checks if the last OpenCL command was successfully executed
//If not, an eroor message is printed
void check_error(int line)
{
	if (g_err != CL_SUCCESS)
		print_error(line - 1, get_error_string(g_err));
}

//source: http://tom.scogland.com/blog/2013/03/29/opencl-errors/
const char *get_error_string(cl_int err)
{
	switch (err) {
		 case 0: return "CL_SUCCESS";
		 case -1: return "CL_DEVICE_NOT_FOUND";
		 case -2: return "CL_DEVICE_NOT_AVAILABLE";
		 case -3: return "CL_COMPILER_NOT_AVAILABLE";
		 case -4: return "CL_MEM_OBJECT_ALLOCATION_FAILURE";
		 case -5: return "CL_OUT_OF_RESOURCES";
		 case -6: return "CL_OUT_OF_HOST_MEMORY";
		 case -7: return "CL_PROFILING_INFO_NOT_AVAILABLE";
		 case -8: return "CL_MEM_COPY_OVERLAP";
		 case -9: return "CL_IMAGE_FORMAT_MISMATCH";
		 case -10: return "CL_IMAGE_FORMAT_NOT_SUPPORTED";
		 case -11: return "CL_BUILD_PROGRAM_FAILURE";
		 case -12: return "CL_MAP_FAILURE";

		 case -30: return "CL_INVALID_VALUE";
		 case -31: return "CL_INVALID_DEVICE_TYPE";
		 case -32: return "CL_INVALID_PLATFORM";
		 case -33: return "CL_INVALID_DEVICE";
		 case -34: return "CL_INVALID_CONTEXT";
		 case -35: return "CL_INVALID_QUEUE_PROPERTIES";
		 case -36: return "CL_INVALID_COMMAND_QUEUE";
		 case -37: return "CL_INVALID_HOST_PTR";
		 case -38: return "CL_INVALID_MEM_OBJECT";
		 case -39: return "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR";
		 case -40: return "CL_INVALID_IMAGE_SIZE";
		 case -41: return "CL_INVALID_SAMPLER";
		 case -42: return "CL_INVALID_BINARY";
		 case -43: return "CL_INVALID_BUILD_OPTIONS";
		 case -44: return "CL_INVALID_PROGRAM";
		 case -45: return "CL_INVALID_PROGRAM_EXECUTABLE";
		 case -46: return "CL_INVALID_KERNEL_NAME";
		 case -47: return "CL_INVALID_KERNEL_DEFINITION";
		 case -48: return "CL_INVALID_KERNEL";
		 case -49: return "CL_INVALID_ARG_INDEX";
		 case -50: return "CL_INVALID_ARG_VALUE";
		 case -51: return "CL_INVALID_ARG_SIZE";
		 case -52: return "CL_INVALID_KERNEL_ARGS";
		 case -53: return "CL_INVALID_WORK_DIMENSION";
		 case -54: return "CL_INVALID_WORK_GROUP_SIZE";
		 case -55: return "CL_INVALID_WORK_ITEM_SIZE";
		 case -56: return "CL_INVALID_GLOBAL_OFFSET";
		 case -57: return "CL_INVALID_EVENT_WAIT_LIST";
		 case -58: return "CL_INVALID_EVENT";
		 case -59: return "CL_INVALID_OPERATION";
		 case -60: return "CL_INVALID_GL_OBJECT";
		 case -61: return "CL_INVALID_BUFFER_SIZE";
		 case -62: return "CL_INVALID_MIP_LEVEL";
		 case -63: return "CL_INVALID_GLOBAL_WORK_SIZE";
		 default: return "Unknown OpenCL error";
		}
}
