#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable


//Uses a desriptor on the amino acid sequences and normalizes them to one length
//aa_descriptor: Values for the amino acids from A to Y
//	If the character doesn't describe an acid, the value is 0
//normalized_length: The length the sequences are to be interpolated
//no_of_sequences
//sequence_start_at: At which position each sequence begins
//sequences: All sequences in one string without any delimiter
//encoded: The sequences after the descriptor has been used
//result: Encoded normalized to one length. Rows are attributes, cols sequences
#define USE_LOCAL_DESCRIPTOR 0

__kernel void interpol_AAdescriptor(__global float aa_descriptor[],
	int normalized_length, int no_of_sequences, __global int sequence_starts_at[],
	__global char sequences[], __global float result[])
{
	#if USE_LOCAL_DESCRIPTOR
	__local float lc_descriptor[CHAR_MAX];
	for (int i = get_local_id(0); i < CHAR_MAX; i += get_local_size(0))
		lc_descriptor[i] = aa_descriptor[i];
	#endif
	
	#if USE_LOCAL_DESCRIPTOR
	for (int seq_index = get_group_id(0); seq_index < no_of_sequences;
			seq_index  += get_num_groups(0)) {
	#else
	int seq_index = get_group_id(0);
	#endif
	
	__global char *our_sequence = sequences + sequence_starts_at[seq_index];
	__global float *our_result = result + seq_index;
	
	int our_length = sequence_starts_at[seq_index + 1]
					- sequence_starts_at[seq_index];
	float factor = (float)(our_length - 1)/ (normalized_length - 1);
	__local int is_invalid;
	if (get_local_id(0) == 0) {
		is_invalid = false;
	}

	barrier(CLK_LOCAL_MEM_FENCE);
	
	//Interpolate
	//Same length. Just encode
	if (our_length == normalized_length) {
		for (int aa = get_local_id(0); aa < normalized_length;
				aa += get_local_size(0)) {
			#if USE_LOCAL_DESCRIPTOR
			float result = lc_descriptor[our_sequence[aa]];
			#else
			float result = aa_descriptor[our_sequence[aa]];
			#endif
			if (isnan(result))
				atom_inc(&is_invalid);
			our_result[aa * no_of_sequences] = result;
		}
	} else {
		for (int aa = get_local_id(0); aa < normalized_length;
				aa += get_local_size(0)) {//For every aa in the interpolated sequence
			float left;
			//aa*factor = original posisition; fract(x) = x – floor(x), left = floor(x)
			float rest = fract(aa * factor, &left);
			int left_int = (int)left;//So it doesn't is calculated two times
			#if USE_LOCAL_DESCRIPTOR
			float left_val = lc_descriptor[our_sequence[left_int]];
			#else
			float left_val = aa_descriptor[our_sequence[left_int]];
			#endif
			//mix(x, y, a) == x + (y - x) * a, a in [0.0,1.0]
			float result = rest < 0.001f ? left_val					
					: mix(left_val,
						#if USE_LOCAL_DESCRIPTOR
						lc_descriptor[our_sequence[left_int + 1]],
						#else
						aa_descriptor[our_sequence[left_int + 1]],
						#endif
						rest);
			our_result[aa * no_of_sequences] = result;
			if (isnan(result))
				atom_inc(&is_invalid);
		}
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	if (is_invalid) {
		for (int i = get_local_id(0); i < normalized_length; i += get_local_size(0))
			our_sequence[i * no_of_sequences] = NAN;
	}
	#if USE_LOCAL_DESCRIPTOR
	}
	#endif
}

//A point in 3D space
//For better readable source code
typedef struct {
	float x;
	float y;
	float z;
} pos3D;


//Calculates the distances between the amino acids and the hull points
//Saves the result in dist_mat
//work_group: amino acid
//work_items: iterate through hull points
__kernel void calc_dist_mat(__constant pos3D aa_pos[], __global pos3D hull_pos[],
	int no_of_hull_points, __global float dist_mat[])
{
	pos3D our_aa = aa_pos[get_group_id(0)];
	__global float* our_row = dist_mat + get_group_id(0) * no_of_hull_points;
	for (int hull_point = get_local_id(0); hull_point < no_of_hull_points;
				hull_point += get_local_size(0)) {
		our_row[hull_point] =
				distance((float3)(our_aa.x, our_aa.y, our_aa.z),
							(float3)(hull_pos[hull_point].x, hull_pos[hull_point].y,
							hull_pos[hull_point].z));
	}
}

//Calculates the field strength at the position of the hull points
//q_mat: charges of the aa. Row aa, col sequence
//dist_mat: distance between hull points and aa. Row aa, col hull point
//e_mat: the result. Row hull point, col sequence
__kernel void calc_strength(__global float q_mat[],
	__global float dist_mat[], 
	int no_of_aa, int no_of_hull_points,
	__global float e_mat[])
{
	//get_group_id(0) <=> sequence number
	//get_num_groups(0) <=> no of sequences
	for (int hull_point = get_local_id(0); hull_point < no_of_hull_points;
				hull_point += get_local_size(0)) {
		__global float *my_q_mat = q_mat + get_group_id(0);
		__global float *my_dist_mat = dist_mat + hull_point;
		float esp = 0;
		for (int aa = 0; aa < no_of_aa; aa++) {
			esp += (*my_q_mat) / (*my_dist_mat);
			my_q_mat += get_num_groups(0);
			my_dist_mat += no_of_hull_points;
		}
		e_mat[hull_point * get_num_groups(0) + get_group_id(0)] = esp;
	}
}

//In inner nodes the left half of data saves the left child's index
//	and the right half the rule's attribute.
//In leafs data saves the class and rule_threshold must be NaN
typedef struct {
	unsigned int data;
	float rule_threshold;
} node;

#define GET_LEFT_CHILD(data) (data >> 16)
#define GET_RULE_ATTRIBUTE(data) (data & USHRT_MAX)

//Builds nodes from information given by R
//trees: Array of Nodes. Rows are trees, cols are nodes
//rest: Rows are trees, cols are nodes
__kernel void build_nodes(int no_of_nodes_per_tree,
	__global int left_children[], 
	__global int bestvar[], __global float xbestsplit[], __global int nodepred[],
	__global node trees[])
{
	//position of the tree's root, which is calculated by this work-group
	int root_index =  get_group_id(0) * no_of_nodes_per_tree;
	//Iterate step-wise
	for (int node_index = get_local_id(0); node_index < no_of_nodes_per_tree; 
			node_index += get_local_size(0)) {
		//The position of the currently calculated node
		int current_index = root_index + node_index;
		//Pointer to the current node
		__global node *current_node = trees + current_index;
		if(left_children[current_index] > 0) {
			current_node->data = left_children[current_index] - 1;
			current_node->data <<= 16;
			current_node->data |= (unsigned int) bestvar[current_index] - 1;
			current_node->rule_threshold = xbestsplit[current_index];
		} else {
			current_node->data = nodepred[current_index];
			current_node->rule_threshold = NAN;
		}
	}
}

//Classifies all sequences on one tree
//trees: Array of Nodes. Rows are trees, cols are nodes
//sequences: Rows are attributes, cols are sequences
//classification: Array which saves for each sequence how many trees classified into class 2
//our_tree: The tree fits into the local memory so it's saved there
#define MAX_LOCAL_INDEX 1024
__kernel void classify_on_one_tree_in_local_memory (
	int no_of_sequences, int no_of_nodes_per_tree,
	__global float sequences[], __global node trees[],
	__global int classification[],
	__local node our_tree[])
{
	//Pointer to the here used tree's root
	bool in_local = 1;
	__global node* root = trees + get_group_id(0) * no_of_nodes_per_tree;
	int copy_until = min(no_of_nodes_per_tree, MAX_LOCAL_INDEX + 1);
	for (int tree_index = get_local_id(0); tree_index < copy_until;
			tree_index += get_local_size(0)) {
		our_tree[tree_index] = root[tree_index];
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	
	//Iterate step-wise	
	for (int sequence_index = get_local_id(0); sequence_index < no_of_sequences;
			sequence_index += get_local_size(0)) {
		int address = 0;
		if (isnan(sequences[GET_RULE_ATTRIBUTE(our_tree[address].data)
							 * no_of_sequences + sequence_index])) {
			atom_xchg(classification + sequence_index, INT_MIN);
			continue;
	 	}
		
		//Use rule contained by node on the sequence to estimate the next node
		//Until the nodes becomes a leaf
		while (!isnan(our_tree[address].rule_threshold)) {
			address = GET_LEFT_CHILD(our_tree[address].data)
			+ (
				sequences[GET_RULE_ATTRIBUTE(our_tree[address].data)
					* no_of_sequences + sequence_index
				] > our_tree[address].rule_threshold
			);
			if (address > MAX_LOCAL_INDEX) {
				in_local = 0;
				break;
			}
		}
		if (!in_local) {
			while (!isnan(root[address].rule_threshold)) {
				address = GET_LEFT_CHILD(root[address].data)
				+ (
					sequences[GET_RULE_ATTRIBUTE(root[address].data)
						* no_of_sequences + sequence_index
					] > root[address].rule_threshold
				);
			}
			if (root[address].data == 2)
				atom_inc(classification + sequence_index);
		} else {
			//Multiple work_items can classify a sequence concurrently so use an atomic instruction
			if (our_tree[address].data == 2)
				atom_inc(classification + sequence_index);
		}
	}
}

__kernel void classify_on_one_tree_in_global_memory (
	int no_of_sequences, int no_of_nodes_per_tree,
	__global float sequences[], __global node trees[],
	__global int classification[])
{
	//Pointer to the here used tree's root
	__global node* root = trees + get_group_id(0) * no_of_nodes_per_tree;
	
	//Iterate step-wise	
	for (int sequence_index = get_local_id(0); sequence_index < no_of_sequences;
			sequence_index += get_local_size(0)) {
		int address = 0;
		
		//Use rule contained by node on the sequence to estimate the next node
		//Until the nodes becomes a leaf
		while (!isnan(root[address].rule_threshold)) {
			address = GET_LEFT_CHILD(root[address].data)
			+ (
				sequences[GET_RULE_ATTRIBUTE(root[address].data)
					* no_of_sequences + sequence_index
				] > root[address].rule_threshold
			);
		}
		
		//Multiple work_items can classify a sequence concurrently so use an atomic instruction
		if (root[address].data == 2)
			atom_inc(classification + sequence_index);
	}
}
