#include <time.h>
//Which print function is to use
#ifndef PROFILING
#	define my_print(...) Rprintf(__VA_ARGS__)
#else
#	define my_print(...) printf(__VA_ARGS__)
#endif

#define CHK_ERR() check_error(__LINE__);

//The name of the file which contains the OpenCL source code
#define CL_FILE_NAME "ClassifyOnGPU.cl"

#include <CL/cl.h>	//OpenCL
#include <R.h>
#include <Rinternals.h>

//For the reading of the OpenCL source code from a file
#include <stdio.h>
#include <errno.h>
#include <string.h>

//Needed for sizeof command
//In inner nodes the left half of data saves the left child's index
//	and the right half the rule's attribute.
//In leafs data saves the class and rule_threshold must be NaN
typedef struct {
	unsigned int data;
	float rule_threshold;
} node;

void check_error(int line);
const char *get_error_string(cl_int err);
void print_error(int line, const char *err_msg);
char *read_file(char *file_name);

float *double_to_float(SEXP from_dbl);
void call_everything(
	int no_of_sequences, char sequences[], int starts_at[],
	float descriptor[], int normalized_length,
	int no_of_trees, int no_of_nodes_per_tree, int left_children[],
		int bestvar[], float xbestsplit[], int nodepred[],
	_Bool calculate_ESP, int no_of_hullP, float aa_pos[], float hull_pos[],
	double result[]
);
cl_mem interpol_AAdescriptor_on_GPU(float aa_descriptor[], int normalized_length,
									int no_of_sequences, int starts_at[],
									char sequences[]);
cl_mem calc_dist_mat(int no_of_aa, int no_of_hullP,
						float aa_pos[], float hull_pos[]);
cl_mem calc_strength_on_GPU(int no_of_sequences, int no_of_aa, int no_of_hullP,
							cl_mem charges_mem, cl_mem dist_mat_mem);
cl_mem build_nodes(int no_of_trees, int no_of_nodes_per_tree,
				int left_children[],
				int bestvar[], float xbestsplit[], int nodepred[]
				);
cl_mem classify_on_GPU(int no_of_sequences, int no_of_attributes,
						int no_of_trees, int no_of_nodes_per_tree,
						cl_mem sequences_mem, cl_mem nodes_mem);

cl_int g_err;
cl_platform_id g_platform;
cl_device_id g_device;
cl_context g_context;
cl_command_queue g_command_queue;
cl_program g_program;

//Executed when library is loaded by R
//Initializes ressources needed by OpenCL
//For the creation of the program the path of the source file is needed
void init(char *package_path[])
{
	//Find available platforms
	cl_uint no_of_platforms;
	g_err = clGetPlatformIDs(0, NULL, &no_of_platforms);
	if (no_of_platforms == 0){
		error("Couldn't find any OpenCL platform");
	}
	CHK_ERR();
	cl_platform_id *platforms = (cl_platform_id *)
				malloc(no_of_platforms * sizeof(cl_platform_id));
	g_err = clGetPlatformIDs(no_of_platforms, platforms, NULL);
	CHK_ERR();

	//Look for a GPU
	_Bool found_gpu = FALSE;
	for (int i = 0; !found_gpu && i < no_of_platforms; i++) {
		cl_uint no_of_gpus;
		clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &no_of_gpus);
		if (no_of_gpus != 0) {
			g_platform = platforms[i];
			found_gpu = TRUE;
		}
	}
	if (!found_gpu) {
		my_print("Couldn't find any GPU. The first OpenCL platform will be used.\n");
		g_platform = platforms[0];
	}

	free(platforms);

	//Get device
	g_err = clGetDeviceIDs(g_platform, CL_DEVICE_TYPE_ALL, 1, &g_device, NULL);
	CHK_ERR();

	//Create context
	g_context = clCreateContext(NULL, 1, &g_device, NULL, NULL, &g_err);
	CHK_ERR();

	//Create command queue
	g_command_queue = clCreateCommandQueue(g_context, g_device, 0, &g_err);
	CHK_ERR();

	//Create Program
	char file_path[1000];
	const char* file_content;

	file_path[0] = '\0';
	#ifndef PROFILING
	strcat(file_path, package_path[0]);
	#else
	strcat(file_path, "../inst");
	#endif
	strcat(file_path, "/OpenCL/");
	strcat(file_path, CL_FILE_NAME);
	file_content = read_file(file_path);

	g_program = clCreateProgramWithSource(g_context, 1, &file_content,
										NULL, &g_err);
	CHK_ERR();

	//Compile OpenCL source code and print it, if there are errors
	g_err = clBuildProgram(g_program, 0, NULL, NULL, NULL, NULL);
	CHK_ERR();
	if (g_err != CL_SUCCESS) {
		char *build_log;
		size_t build_log_size;
		g_err = clGetProgramBuildInfo(g_program, g_device, CL_PROGRAM_BUILD_LOG,
										0, NULL, &build_log_size);
		CHK_ERR();
		build_log = (char *)malloc(build_log_size);
		g_err = clGetProgramBuildInfo(g_program, g_device, CL_PROGRAM_BUILD_LOG,
										build_log_size, build_log, NULL);
		CHK_ERR();
		my_print("%s\n", build_log);
		free((void *)build_log);
	}

	free((void *)file_content);
}

//Executed when library is unloaded by R
//Deletes all created OpenCL objects
void R_unload_ClassifyOnGPU()
{
	g_err = clReleaseCommandQueue(g_command_queue);
	CHK_ERR();

	g_err = clRetainProgram(g_program);
	CHK_ERR();

	g_err = clReleaseContext(g_context);
	CHK_ERR();
}

//Transforms a array of doubles into an array of floats
//Needed beacause R doesn't support single precision floats
//Don't forget free
float *double_to_float(SEXP from_dbl){
	int len = length(from_dbl);
	float *to = (float *)malloc(sizeof(float) * len);
	double *from = REAL(from_dbl);
	for(int i = 0; i < len; i++)
		to[i] = (float)from[i];
	return to;
}

//Called by R
//sequences: Array of strings which represent the sequences
//decriptor: Array which contains the values for the encoding of amino acids
//normalized_length: The length the encoded sequences are to be interpolated to
//no_of_trees - node_pred: The random forest
//aa_pos: The positions of the amino acids in the HIV-structure
//hull_pos: The hull points' positions in which the ESP will be calculated
SEXP use_GPU_for_classification(
	SEXP sequences,
	SEXP descriptor, SEXP normalized_length,
	SEXP no_of_trees, SEXP left_children, SEXP bestvar,
		SEXP xbestsplit, SEXP nodepred,
	SEXP aa_pos, SEXP hull_pos
)
{
	//Calculation of ESP needed?
	_Bool calc_esp = !isNull(aa_pos);

	//sequences
	int i, len;
	int no_of_sequences = length(sequences);
	//Where the sequences in the concatenate string start
	int *starts_at = (int *)malloc(sizeof(int) * (no_of_sequences + 1));
	char *all_sequences, *copy_to;
	starts_at[0] = 0;
	for (i = 1; i <= no_of_sequences; i++){
		starts_at[i] = starts_at[i - 1] + strlen(CHAR(STRING_ELT(sequences, i - 1)));
	}
	all_sequences = (char *)malloc(sizeof(char) * starts_at[no_of_sequences]);
	copy_to = all_sequences;
	for(i = 0; i < no_of_sequences; i++){
		len = starts_at[i + 1] - starts_at[i];
		strncpy(copy_to, CHAR(STRING_ELT(sequences, i)), len);
		copy_to += len;
	}

	//descriptor, xbestsplit, pos
	//transform arrays of doubles into arrays of floats
	float *descriptor_flt = double_to_float(descriptor);
	float *xbestsplit_flt = double_to_float(xbestsplit);
	float *aa_pos_flt = NULL, *hull_pos_flt = NULL;
	if (calc_esp) {
		aa_pos_flt = double_to_float(aa_pos);
		hull_pos_flt = double_to_float(hull_pos);
	}

	//result
	SEXP result;
	PROTECT(result = allocVector(REALSXP, no_of_sequences));

	int tree_quantity = INTEGER(no_of_trees)[0];
	//Classify on the GPU
	call_everything(no_of_sequences, all_sequences, starts_at,
					descriptor_flt, INTEGER(normalized_length)[0],
					tree_quantity, length(left_children) / tree_quantity,
						INTEGER(left_children),
						INTEGER(bestvar), xbestsplit_flt, INTEGER(nodepred),
					calc_esp, calc_esp ? length(hull_pos) / 3 : 0,
						aa_pos_flt, hull_pos_flt,
					REAL(result));

	free(starts_at);
	free(all_sequences);
	free(descriptor_flt);
	free(xbestsplit_flt);
	if (calc_esp) {
		free(aa_pos_flt);
		free(hull_pos_flt);
	}
	UNPROTECT(1);//Give the control over result to the R garbage collector
	return result;
}

//Calls the single kernels and transfers the results between them
void call_everything(
	int no_of_sequences, char sequences[], int starts_at[],
	float descriptor[], int normalized_length,
	int no_of_trees, int no_of_nodes_per_tree, int left_children[],
		int bestvar[], float xbestsplit[], int nodepred[],
	_Bool calculate_ESP, int no_of_hullP, float aa_pos[], float hull_pos[],
	double result[]
)
{
	cl_mem encoded_mem, esp_vals_mem, nodes_mem, result_mem;
	encoded_mem = interpol_AAdescriptor_on_GPU(descriptor, normalized_length,
					no_of_sequences, starts_at, sequences);
	if (calculate_ESP) {
		//normalized_length = no_of_aa
		cl_mem dist_mat_mem = calc_dist_mat(normalized_length, no_of_hullP,
								aa_pos, hull_pos);
		esp_vals_mem = calc_strength_on_GPU(no_of_sequences, normalized_length,
						no_of_hullP, encoded_mem, dist_mat_mem);
		g_err = clReleaseMemObject(dist_mat_mem);
		g_err |= clReleaseMemObject(encoded_mem);
		CHK_ERR();
	}

	nodes_mem = build_nodes(no_of_trees, no_of_nodes_per_tree, left_children,
					bestvar, xbestsplit, nodepred);
	result_mem = classify_on_GPU(no_of_sequences,
					calculate_ESP ? no_of_hullP : normalized_length,
					no_of_trees, no_of_nodes_per_tree,
					calculate_ESP ? esp_vals_mem : encoded_mem, nodes_mem);

	if (calculate_ESP) {
		g_err = clReleaseMemObject(esp_vals_mem);
		CHK_ERR();
	} else {
		g_err = clReleaseMemObject(encoded_mem);
		CHK_ERR();
	}
	g_err = clReleaseMemObject(nodes_mem);
	CHK_ERR();

	int *classification = (int *)malloc(sizeof(int) * no_of_sequences);
	g_err = clEnqueueReadBuffer(g_command_queue, result_mem, CL_TRUE, 0,
				no_of_sequences * sizeof(int), classification, 0, NULL, NULL);
	CHK_ERR();
	g_err = clReleaseMemObject(result_mem);
	CHK_ERR();

	//So it must be converted only once
	double no_of_trees_dbl = (double)no_of_trees;
	for(int i = 0; i < no_of_sequences; i++){
		result[i] = classification[i] < 0 ? NAN : classification[i] / no_of_trees_dbl;
	}
	free(classification);
}

//Uses a descriptor on the amino acids and interpolates them to one length
//aa_descriptor: Values for the amino acids from A to Y
//	If the character doesn't describe an acid, the value is 0
//normalized_length: The length the sequences are to be interpolated
//no_of_sequences
//starts_at: Where each sequence starts
//sequences: All sequences in one string without any delimiter
//result: Encoded and normalized sequences
//	Rows are attributes, cols are sequences
cl_mem interpol_AAdescriptor_on_GPU(float aa_descriptor[], int normalized_length,
									int no_of_sequences, int starts_at[],
									char sequences[])
{
	cl_kernel kernel;
	cl_mem aa_descriptor_mem, starts_at_mem, sequences_mem, result_mem;
	size_t work_items_per_work_group = 192;

	//size_t total_work_items = work_items_per_work_group * 1024;
	size_t total_work_items = work_items_per_work_group * no_of_sequences;

	//Allocate memory
	aa_descriptor_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
						CHAR_MAX * sizeof(float), aa_descriptor, &g_err);
	CHK_ERR();

	starts_at_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					(no_of_sequences + 1) * sizeof(int), starts_at, &g_err);
	CHK_ERR();

	sequences_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					starts_at[no_of_sequences] * sizeof(char), sequences, &g_err);
	CHK_ERR();

	result_mem = clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					no_of_sequences * normalized_length * sizeof(float), NULL, &g_err);
	CHK_ERR();

	//Create kernel
	kernel = clCreateKernel(g_program, "interpol_AAdescriptor", &g_err);
	CHK_ERR();

	//Set parameters
	g_err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &aa_descriptor_mem);
	g_err |= clSetKernelArg(kernel, 1, sizeof(int), &normalized_length);
	g_err |= clSetKernelArg(kernel, 2, sizeof(int), &no_of_sequences);
	g_err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), &starts_at_mem);
	g_err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), &sequences_mem);
	g_err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), &result_mem);
	CHK_ERR();

	//Run
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
			&total_work_items, &work_items_per_work_group, 0, NULL, NULL);
	CHK_ERR();

	//Clean up
	g_err = clReleaseKernel(kernel);
	g_err |= clReleaseMemObject(aa_descriptor_mem);
	g_err |= clReleaseMemObject(starts_at_mem);
	g_err |= clReleaseMemObject(sequences_mem);
	CHK_ERR();

	//Return result
	return result_mem;
}

//Classifies the given sequences on the given nodes
//no_of_attributes: Of how many attributes a sequence consists
//no_of_trees: Oh how many trees the random forest consits
//no_of_nodes_per_tree: The maximal number of nodes a tree consists of
//sequences: Rows are attributes, cols are sequences
//nodes_mem: The nodes the trees consist of. Rows are trees, cols are nodes
//return: For each sequence the number of trees, which voted for class 2
#define MAX_NUMBER_OF_NODES_IN_LOCAL_MEMORY 1024;
cl_mem classify_on_GPU(int no_of_sequences, int no_of_attributes,
						int no_of_trees, int no_of_nodes_per_tree,
						cl_mem sequences_mem, cl_mem nodes_mem)
{
	cl_kernel kernel;
	cl_mem classifications_mem;
	size_t work_items_per_work_group = 192;
	size_t total_work_items = no_of_trees * work_items_per_work_group;

	_Bool use_local_memory =
			no_of_nodes_per_tree <= MAX_NUMBER_OF_NODES_IN_LOCAL_MEMORY;
	const char *kernel_name =	use_local_memory ?
								"classify_on_one_tree_in_local_memory" :
								"classify_on_one_tree_in_global_memory";
	//Initialize memory,
	//	because it's used by all work_groups and so no synchronisation is possible
	int *nulls = calloc(no_of_sequences, sizeof(int));
	classifications_mem = clCreateBuffer(g_context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
							no_of_sequences * sizeof(int), nulls, &g_err);
	CHK_ERR();

	//Create kernel
	kernel = clCreateKernel(g_program, kernel_name, &g_err);
	CHK_ERR();

	//Set parameters
	int param = 0;
	g_err =  clSetKernelArg(kernel, param++, sizeof(int), &no_of_sequences);
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_nodes_per_tree);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &sequences_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &nodes_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &classifications_mem);
	if (use_local_memory)
		g_err |= clSetKernelArg(kernel, param++, no_of_nodes_per_tree * sizeof(node), NULL);
	CHK_ERR();

	//Start calculation
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, 0,
									&total_work_items,
									&work_items_per_work_group,
									0, NULL, NULL);
	CHK_ERR();

	//Clean up
	free(nulls);
	g_err = clReleaseKernel(kernel);
	CHK_ERR();

	//Return result
	return classifications_mem;
}

//Build nodes on the data given by R
//no_of_trees: Of how many trees the random forest constists of
//no_of_nodes_per_tree: Of how many nodes a tree maximal consists of
//leftchildren - nodepred: From these the nodes are build.
//	Rows are trees, cols are nodes
//Returns a cl_mem objects which contains the build nodes
cl_mem build_nodes(int no_of_trees, int no_of_nodes_per_tree,
				int left_children[],
				int bestvar[], float xbestsplit[], int nodepred[])
{
	cl_kernel kernel;
	cl_mem left_children_mem,
			bestvar_mem, xbestsplit_mem, nodepred_mem,
			nodes_mem;
	size_t matrix_size = no_of_trees * no_of_nodes_per_tree * sizeof(int);
	size_t work_items_per_work_group = 192;
	size_t total_work_items = work_items_per_work_group * no_of_trees;

	//Allocate Memory
	left_children_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
						matrix_size, left_children, &g_err);
	CHK_ERR();
	bestvar_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
						matrix_size, bestvar, &g_err);
	CHK_ERR();
	xbestsplit_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
						matrix_size, xbestsplit, &g_err);
	CHK_ERR();
	nodepred_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
						matrix_size, nodepred, &g_err);
	CHK_ERR();
	nodes_mem = clCreateBuffer(g_context, CL_MEM_READ_WRITE,
						no_of_trees * no_of_nodes_per_tree * sizeof(node), NULL, &g_err);
	CHK_ERR();

	//Create kernel
	kernel = clCreateKernel(g_program, "build_nodes", &g_err);
	CHK_ERR();

	//Set parameters
	g_err =  clSetKernelArg(kernel, 0, sizeof(int), &no_of_nodes_per_tree);
	g_err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &left_children_mem);
	g_err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), &bestvar_mem);
	g_err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), &xbestsplit_mem);
	g_err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), &nodepred_mem);
	g_err |= clSetKernelArg(kernel, 5, sizeof(cl_mem), &nodes_mem);
	CHK_ERR();

	//Start calculation
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
								&total_work_items, &work_items_per_work_group,
						 		0, NULL, NULL);
	CHK_ERR();

	//Clean up
	g_err = clReleaseKernel(kernel);
	g_err |= clReleaseMemObject(left_children_mem);
	g_err |= clReleaseMemObject(bestvar_mem);
	g_err |= clReleaseMemObject(xbestsplit_mem);
	g_err |= clReleaseMemObject(nodepred_mem);
	CHK_ERR();

	//Return result
	return nodes_mem;
}

//charges: A matrix of charges with sequences as rows and the the amino acids
//	as cols
//aa_pos: A matrix of positions with aa as rows and the x,y,z coordinates as cols
//hull_pos: same as aa_pos
//result: A matrix of field strengths with sequences as rows, hull points as cols
//package_path: The location of this R-package.
//	Needed for the finding of the OpenCL source code files.

//Calculates the matrix of the distances between the aa and hull points
//no_of_aa: Of how many the amino acids a normalized sequence consits of
//no_of_hulP: Of how many hull points the ESP will be calculated
//aa_pos: The positions of the amino acids.
//	Rows are amino acids, cols x,y,z coordinates
//hull_pos: See aa_pos
//Return: The cl_mem object which contains the matrix of the distances
#define MAX_NO_OF_WORK_ITEMS_IN_CALC_DIST_MAT 128
cl_mem calc_dist_mat(int no_of_aa, int no_of_hullP,
						float aa_pos[], float hull_pos[])
{
	cl_kernel kernel;
	cl_mem aa_pos_mem, hull_pos_mem, dist_mat_mem;
	size_t work_items_per_work_group = no_of_hullP;
	if (work_items_per_work_group > MAX_NO_OF_WORK_ITEMS_IN_CALC_DIST_MAT)
		work_items_per_work_group = MAX_NO_OF_WORK_ITEMS_IN_CALC_DIST_MAT;
	size_t total_work_items = work_items_per_work_group * no_of_aa;

	//Allocate memory on the GPU
	aa_pos_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					no_of_aa * 3 * sizeof(float), aa_pos, &g_err);
	CHK_ERR();
	hull_pos_mem = clCreateBuffer(g_context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					no_of_hullP * 3 * sizeof(float), hull_pos, &g_err);
	CHK_ERR();
	dist_mat_mem = clCreateBuffer(g_context, CL_MEM_READ_WRITE,
					no_of_aa * no_of_hullP * sizeof(float), NULL, &g_err);
	CHK_ERR();

	//Create kernel
	kernel = clCreateKernel(g_program, "calc_dist_mat", &g_err);
	CHK_ERR();

	//Set parameters
	int param = 0;
	g_err =  clSetKernelArg(kernel, param++, sizeof(cl_mem), &aa_pos_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &hull_pos_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_hullP);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &dist_mat_mem);
	CHK_ERR();

	//Start calculation
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
									&total_work_items,
									&work_items_per_work_group,
									0, NULL, NULL);
	CHK_ERR();

	//Clean up
	g_err = clReleaseKernel(kernel);
	g_err |= clReleaseMemObject(aa_pos_mem);
	g_err |= clReleaseMemObject(hull_pos_mem);
	CHK_ERR();

	//Return result
	return dist_mat_mem;
}
//Calculates the field strength of the sequences' hull point
//no_of_sequences: For how many sequences the hull's ESP has to be calculated
//no_of_aa: Of how many amino acids a normalized sequence consits
//no_of_hulP: Of how many hull points the ESP will be calculated
//charges_mem: The amino acids' charges for every sequence
//	Rows are amino acids, cols sequences
//dist_mat_mem: The distance between the amino acids and the hull points
//	Rows are amino acids, cols hull points
//return: The hull points' ESP for every sequence
//	Rows are hull points, cols are sequences
#define MAX_NO_OF_WORK_ITEMS_IN_ESP_CALC 192
cl_mem calc_strength_on_GPU(int no_of_sequences, int no_of_aa, int no_of_hullP,
							cl_mem charges_mem, cl_mem dist_mat_mem)
{
	cl_mem strength_mem;
	cl_kernel kernel;
	size_t work_items_per_work_group = no_of_hullP;
	if (work_items_per_work_group > MAX_NO_OF_WORK_ITEMS_IN_ESP_CALC)
		work_items_per_work_group = MAX_NO_OF_WORK_ITEMS_IN_ESP_CALC;

	size_t total_no_of_work_items = work_items_per_work_group * no_of_sequences;

	//Allocate memory
	strength_mem = clCreateBuffer(g_context, CL_MEM_WRITE_ONLY,
					no_of_sequences * no_of_hullP * sizeof(float), NULL,&g_err);
	CHK_ERR();

	//Create kernel
	kernel = clCreateKernel(g_program, "calc_strength", &g_err);
	CHK_ERR();

	//Set parameters
	int param = 0;
	g_err =  clSetKernelArg(kernel, param++, sizeof(cl_mem), &charges_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &dist_mat_mem);
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_aa);
	g_err |= clSetKernelArg(kernel, param++, sizeof(int), &no_of_hullP);
	g_err |= clSetKernelArg(kernel, param++, sizeof(cl_mem), &strength_mem);
	CHK_ERR();

	//Start calculation
	g_err = clEnqueueNDRangeKernel(g_command_queue, kernel, 1, NULL,
									&total_no_of_work_items,
									&work_items_per_work_group,
									0, NULL, NULL);
	CHK_ERR();

	//Clean up
	g_err = clReleaseKernel(kernel);
	//g_err |= clReleaseMemObject(charges_mem);
	CHK_ERR();

	//Return result
	return strength_mem;
}

//Returns the file content as a string
//Do not forget to use free
//Approach from http://www.cplusplus.com/reference/cstdio/fread/
char *read_file(char *file_name)
{
	FILE *file;
	long file_length;
	char *file_content;
	size_t read_amount;

	//Open file
	file = fopen(file_name, "r");
	if (file == NULL){
		print_error(__LINE__ - 2, strerror(errno));
		return NULL;
	}

	//Determine file length
	fseek(file, 0, SEEK_END);//Go to the end of the file
	file_length = ftell(file);//Read position in file
	rewind(file);//Go to the beginning of the file

	//Allocate memory
	file_content = (char *)malloc(file_length * sizeof(char));

	//Read content
	read_amount = fread(file_content, sizeof(char), file_length, file);
	if (read_amount != file_length) {
		print_error(__LINE__ - 2, "Read less than should");
		return NULL;
	}


	//The string must be null-terminated, so it's length can be determined
	file_content[file_length - 1] = '\0';
	//Close file
	fclose(file);
	return file_content;
}

//Prints an error message
void print_error(int line, const char *err_msg)
{
	my_print("Error in line %d: %s\n", line, err_msg);
}

//Checks if the last OpenCL command was successfully executed
//If not, an eroor message is printed
void check_error(int line)
{
	if (g_err != CL_SUCCESS)
		print_error(line - 1, get_error_string(g_err));
}

//source: http://tom.scogland.com/blog/2013/03/29/opencl-errors/
const char *get_error_string(cl_int err)
{
	switch (err) {
		 case 0: return "CL_SUCCESS";
		 case -1: return "CL_DEVICE_NOT_FOUND";
		 case -2: return "CL_DEVICE_NOT_AVAILABLE";
		 case -3: return "CL_COMPILER_NOT_AVAILABLE";
		 case -4: return "CL_MEM_OBJECT_ALLOCATION_FAILURE";
		 case -5: return "CL_OUT_OF_RESOURCES";
		 case -6: return "CL_OUT_OF_HOST_MEMORY";
		 case -7: return "CL_PROFILING_INFO_NOT_AVAILABLE";
		 case -8: return "CL_MEM_COPY_OVERLAP";
		 case -9: return "CL_IMAGE_FORMAT_MISMATCH";
		 case -10: return "CL_IMAGE_FORMAT_NOT_SUPPORTED";
		 case -11: return "CL_BUILD_PROGRAM_FAILURE";
		 case -12: return "CL_MAP_FAILURE";

		 case -30: return "CL_INVALID_VALUE";
		 case -31: return "CL_INVALID_DEVICE_TYPE";
		 case -32: return "CL_INVALID_PLATFORM";
		 case -33: return "CL_INVALID_DEVICE";
		 case -34: return "CL_INVALID_CONTEXT";
		 case -35: return "CL_INVALID_QUEUE_PROPERTIES";
		 case -36: return "CL_INVALID_COMMAND_QUEUE";
		 case -37: return "CL_INVALID_HOST_PTR";
		 case -38: return "CL_INVALID_MEM_OBJECT";
		 case -39: return "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR";
		 case -40: return "CL_INVALID_IMAGE_SIZE";
		 case -41: return "CL_INVALID_SAMPLER";
		 case -42: return "CL_INVALID_BINARY";
		 case -43: return "CL_INVALID_BUILD_OPTIONS";
		 case -44: return "CL_INVALID_PROGRAM";
		 case -45: return "CL_INVALID_PROGRAM_EXECUTABLE";
		 case -46: return "CL_INVALID_KERNEL_NAME";
		 case -47: return "CL_INVALID_KERNEL_DEFINITION";
		 case -48: return "CL_INVALID_KERNEL";
		 case -49: return "CL_INVALID_ARG_INDEX";
		 case -50: return "CL_INVALID_ARG_VALUE";
		 case -51: return "CL_INVALID_ARG_SIZE";
		 case -52: return "CL_INVALID_KERNEL_ARGS";
		 case -53: return "CL_INVALID_WORK_DIMENSION";
		 case -54: return "CL_INVALID_WORK_GROUP_SIZE";
		 case -55: return "CL_INVALID_WORK_ITEM_SIZE";
		 case -56: return "CL_INVALID_GLOBAL_OFFSET";
		 case -57: return "CL_INVALID_EVENT_WAIT_LIST";
		 case -58: return "CL_INVALID_EVENT";
		 case -59: return "CL_INVALID_OPERATION";
		 case -60: return "CL_INVALID_GL_OBJECT";
		 case -61: return "CL_INVALID_BUFFER_SIZE";
		 case -62: return "CL_INVALID_MIP_LEVEL";
		 case -63: return "CL_INVALID_GLOBAL_WORK_SIZE";
		 default: return "Unknown OpenCL error";
		}
}
