//Which print function is to use
#ifndef PROFILING
#	define my_print(...) Rprintf(__VA_ARGS__)
#else
#	define my_print(...) printf(__VA_ARGS__)
#endif

#include <R.h>
#include <Rinternals.h>
#include <time.h>//random number generator initialization
#include <ctype.h>

int get_no_of_combinations(const char *sequence, int max_combinations);
void create_combinations(int index, const char *sequence, int *starts_at, char **new_sequences, int *origin);
void create_random_combinations(int index, const char *sequence, int *starts_at, char **new_sequences, int *origin);
inline char* to_lower_in_C (const char *original, char **buffer, int *buffer_size);
inline char* to_upper_in_C(const char *original, char **buffer, int *buffer_size);
#define SEARCHED_CHAR 'C'
#define STOP_SIGN '*'
//Trim a sequence so it starts with its first 'C' and ends with its last
//If the trimmed region is shorter than 30 characters long, replace it with NA_STRING
SEXP filterAndAlign(SEXP sequences){
	SEXP result;
	PROTECT(result = allocVector(STRSXP, length(sequences)));
	
	for(int i = 0; i < length(sequences); i++) {
		const char *sequence = CHAR(STRING_ELT(sequences, i));
		const char *first_position;
		const char *last_position;
		int length;
		
		//strchr gives a pointer to the first appearance of a specified character
		first_position = strchr(sequence, SEARCHED_CHAR);
		if (first_position == NULL) {
			SET_STRING_ELT(result, i, NA_STRING);
		} else {
			last_position = strrchr(sequence, SEARCHED_CHAR);
			length = last_position - first_position + 1;
			if (length < 30) {
				SET_STRING_ELT(result, i, NA_STRING);
			} else {
				//Sequences with stop codon "*" aren't allowed
				const char *stop_position = strchr(first_position, STOP_SIGN);
				if (stop_position != NULL && stop_position < last_position)
					SET_STRING_ELT(result, i, NA_STRING);
				else
					SET_STRING_ELT(result, i, mkCharLen(first_position, length));
			}
		}
	}
	
	UNPROTECT(1);//Give control over result to the garbage collector
	return result;
}

SEXP replaceNucleotideAmbiguityCodes(SEXP sequences, SEXP r_max_combinations){
	//r_result: List which contains r_origin and r_new_sequences
	//r_origin[i]: From which original sequence the (i+1)th sequence int r_new_sequences stems from
	//r_new_sequences: The possible encodings for SEXP sequences
	SEXP r_result, r_origin, r_new_sequences;
	PROTECT(r_result = allocVector(VECSXP, 2));
	//starts_at[i]: Where the (i+1)th sequence's combinations start
	int max_combinations = *INTEGER(r_max_combinations);
	
	int *starts_at = (int *)malloc((length(sequences) + 1) * sizeof(int));
	_Bool *no_of_combinations_greater_threshold = (_Bool *)malloc(length(sequences) * sizeof(_Bool));
	
	starts_at[0] = 0;
	for(int i = 0; i < length(sequences); i++) {
		int no_of_combinations = get_no_of_combinations(CHAR(STRING_ELT(sequences, i)), max_combinations);
		no_of_combinations_greater_threshold[i] = no_of_combinations > max_combinations;
		starts_at[i + 1] = starts_at[i] + (no_of_combinations_greater_threshold[i] ? max_combinations : no_of_combinations);
	}
	
	//The possible sequences' encodings
	char **new_sequences = (char **)malloc(starts_at[length(sequences)] * sizeof(char *));
	PROTECT(r_origin = allocVector(INTSXP, starts_at[length(sequences)]));
	int *origin = INTEGER(r_origin);
	//random number generator initialization
	srand(time(NULL));
	
	//Create encodings for each original sequence
	for (int  i = 0; i < length(sequences); i++) {
		if (no_of_combinations_greater_threshold[i])
			create_random_combinations(i, CHAR(STRING_ELT(sequences, i)), starts_at, new_sequences, origin);
		else
			create_combinations(i, CHAR(STRING_ELT(sequences, i)), starts_at, new_sequences, origin);
	}
	
	PROTECT(r_new_sequences = allocVector(STRSXP, starts_at[length(sequences)]));

	//Create R-strings and free memory, which was allocated by create_combinations
	for(int i = 0; i < starts_at[length(sequences)]; i++){
		SET_STRING_ELT(r_new_sequences, i, mkChar(new_sequences[i]));
		free(new_sequences[i]);
	}
	
	free(starts_at);
	free(new_sequences);
	
	SET_VECTOR_ELT(r_result, 0, r_origin);
	SET_VECTOR_ELT(r_result, 1, r_new_sequences);
	UNPROTECT(3);//Give control over objects to the R-garbage collector
	return r_result;
}


//How many possible encodings for each character exist
//not specified values are 0
static int no_of_variants[128] =
{
	['a'] = 1,//a
	['b'] = 3,//b
	['c'] = 1,//c
	['d'] = 3,//d
	['g'] = 1,//g
	['h'] = 3,//h
	['k'] = 2,//k
	['m'] = 2,//m
	['n'] = 4,//n
	['r'] = 2,//r
	['s'] = 2,//s
	['t'] = 1,//t
	['v'] = 3,//v
	['w'] = 2,//w
	['y'] = 2,//y
};

//The possible encodings
static char variants[128][4] =
{
	['a'] = {'a'},//a
	['b'] = {'g','t','c'},//b
	['c'] = {'c'},//c
	['d'] = {'g', 'a', 't'},//d
	['g'] = {'g'},//g
	['h'] = {'a', 'c', 't'},//h
	['k'] = {'g', 't'},//k
	['m'] = {'a', 'c'},//m
	['n'] = {'g', 'a', 't', 'c'},//n
	['r'] = {'g', 'a'},//r
	['s'] = {'g', 'c'},//s
	['t'] = {'t'},//t
	['v'] = {'g', 'c', 'a'},//v
	['w'] = {'a', 't'},//w
	['y'] = {'t', 'c'},//y
};

//returns the number of possible encodings
//	if this number is greater max_combinations, it returns max_combinations + 1
int get_no_of_combinations(const char *sequence, int max_combinations)
{
	int no_of_combinations = 1;
	for (const char *iter = sequence; *iter != '\0'; iter++) {
		int current_variants = no_of_variants[(int)*iter];
		//==0 if this a character, which doesn't describe one or multiple nucleobases
		if (current_variants == 0)
			error("The following sequence contains a character that cannot be resolved: %s", sequence);	
		no_of_combinations *= current_variants;
		//Check every time so no integer overflow appears
		if (no_of_combinations > max_combinations)
			return max_combinations + 1;
	}	
	return no_of_combinations;
}

void create_combinations(int index, const char *sequence, int *starts_at, char **new_sequences, int *origin)
{
	//Where the strings are, which will be written here
	int my_start = starts_at[index];
	int no_of_combinations = starts_at[index + 1] - my_start;
	//Allocate memory for the strings and null terminate them
	for(int i = my_start; i < my_start + no_of_combinations; i++) {
		new_sequences[i] = (char *)malloc((strlen(sequence) + 1) * sizeof(char));
		new_sequences[i][strlen(sequence)] = '\0';
		origin[i] = index + 1;
	}
	
	if (no_of_combinations == 1) {
		strcpy(new_sequences[my_start], sequence);
		return;
	}
	//block: an amount of strings that are equal
	int no_of_current_combinations = 1;
	int current_block_height = no_of_combinations;
	//Iterate through characters and place them in each of the new sequences
	for(int c = 0; c < strlen(sequence); c++) {
		int current_variants = no_of_variants[(int)sequence[c]];
		//Else-case code works also for current_variants == 1 but needs more calculations
		if (current_variants == 1) {
			char used_variant = variants[(int)sequence[c]][0];
			for(int i = 0; i < no_of_combinations; i++) {
				new_sequences[my_start + i][c] = used_variant;
			}
		} else {
			int new_block_height = current_block_height / current_variants;
			for(int i = 0; i < no_of_current_combinations; i++){
				for (int j = 0; j < current_variants; j++) {
					char used_variant = variants[(int)sequence[c]][j];
					for (int k = i * current_block_height + j * new_block_height; k < i * current_block_height + (j + 1) * new_block_height; k++){
						new_sequences[my_start + k][c] = used_variant;					
					}
				}
			}
			no_of_current_combinations *= current_variants;
			current_block_height /= current_variants;	
		}
	}

}

void create_random_combinations(int index, const char *sequence, int *starts_at, char **new_sequences, int *origin)
{
	//Where the strings are, which will be written here
	int my_start = starts_at[index];
	int no_of_combinations = starts_at[index + 1] - my_start;//== max_combinations
	//Allocate memory for the strings and null terminate them
	for(int i = my_start; i < my_start + no_of_combinations; i++) {
		new_sequences[i] = (char *)malloc((strlen(sequence) + 1) * sizeof(char));
		new_sequences[i][strlen(sequence)] = '\0';
		origin[i] = index + 1;
	}

	//Iterate through characters and place them in each of the new sequences
	for(int c = 0; c < strlen(sequence); c++) {
		int current_variants = no_of_variants[(int)sequence[c]];
		//Else-case code works also for current_variants == 1 but needs more calculations
		if (current_variants == 1) {
			char used_variant = variants[(int)sequence[c]][0];
			for(int i = 0; i < no_of_combinations; i++) {
				new_sequences[my_start + i][c] = used_variant;
			}
		} else {
			for(int i = 0; i < no_of_combinations; i++) {
				new_sequences[my_start + i][c] = variants[(int)sequence[c]][rand()%current_variants];
			}
		}
	}

}

//Calculate maximal classification result for the orginal sequences
SEXP getMaxima(SEXP r_values, SEXP r_origin, SEXP r_no_of_original_sequences)
{
	SEXP r_result;
	int no_of_original_sequences = *INTEGER(r_no_of_original_sequences);
	PROTECT(r_result = allocVector(REALSXP, no_of_original_sequences));
	double *result = REAL(r_result);
	double *values = REAL(r_values);
	int *origin = INTEGER(r_origin);
	//Initialize with a negative value so one can see if the values was set
	for (int i = 0; i < no_of_original_sequences; i++) {
		result[i] = -1.0;
	}
	for (int i = 0; i < length(r_values); i++) {
		if (!isnan(values[i])) {
			result[origin[i] - 1] = fmax(result[origin[i] - 1], values[i]);
		}
	}
	//If the value is < 0, it wasn't set. Set it to NAN
	for (int i = 0; i < no_of_original_sequences; i++) {
		if (result[i] < 0)
			result[i] = NAN;
	}
	UNPROTECT(1);
	return r_result;
}

SEXP to_lower(SEXP sequences)
{
	int no_of_sequences = length(sequences);
	SEXP r_result;
	PROTECT(r_result = allocVector(STRSXP, no_of_sequences));
	
	int buffer_size = 1000;
	char *buffer = (char*)malloc(buffer_size * sizeof(char));
	
	for (int i = 0; i < no_of_sequences; i++) {
		to_lower_in_C(CHAR(STRING_ELT(sequences, i)), &buffer, &buffer_size);
		SET_STRING_ELT(r_result, i, mkChar(buffer));
	}
	
	free(buffer);
	UNPROTECT(1);
	return r_result;
}

inline char* to_lower_in_C(const char *original, char **buffer, int *buffer_size)
{
	int char_length = strlen(original);
	if (char_length > *buffer_size) {
		free(*buffer);
		*buffer_size = char_length;
		*buffer = (char *)malloc((*buffer_size + 1) * sizeof(char));
	}
	for (int i = 0; i <= char_length; i++)
		(*buffer)[i] = tolower(original[i]);
}

SEXP to_upper(SEXP sequences)
{
	int no_of_sequences = length(sequences);
	SEXP r_result;
	PROTECT(r_result = allocVector(STRSXP, no_of_sequences));

	int buffer_size = 1000;
	char *buffer = (char*)malloc(buffer_size * sizeof(char));

	for (int i = 0; i < no_of_sequences; i++) {
		to_upper_in_C(CHAR(STRING_ELT(sequences, i)), &buffer, &buffer_size);
		SET_STRING_ELT(r_result, i, mkChar(buffer));
	}

	free(buffer);
	UNPROTECT(1);
	return r_result;
}

inline char* to_upper_in_C(const char *original, char **buffer, int *buffer_size)
{
	int char_length = strlen(original);
	if (char_length > *buffer_size) {
		free(*buffer);
		*buffer_size = char_length;
		*buffer = (char *)malloc((*buffer_size + 1) * sizeof(char));
	}
	for (int i = 0; i <= char_length; i++)
		(*buffer)[i] = toupper(original[i]);
}
